/**
 * Copyright (c) 2021 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "assetmanager.h"

assetmanagerloaderdefinition_t ASSET_MANAGER_LOADERS[] = {
  {
    &_assetManagerLoaderTextureAsync,
    &_assetManagerLoaderTextureSync
  },
  {
    &_assetManagerLoaderFontAsync,
    &_assetManagerLoaderFontSync
  },
  {
    &_assetManagerLoaderShaderAsync,
    &_assetManagerLoaderShaderSync
  }
};


void assetManagerInit(assetmanager_t *manager) {
  threadInit(&manager->thread, &_assetManagerThread);
  manager->thread.user = manager;
  manager->itemCount = 0;
  manager->finished = false;
}

void assetManagerStart(assetmanager_t *manager) {
  threadStart(&manager->thread);
}

float assetManagerProgressGet(assetmanager_t *manager) {
  float done;
  uint8_t i;

  done = 0.0f;

  for(i = 0; i < manager->itemCount; i++) {
    if(!assetManagerIsItemFinished(manager->items + i)) continue;
    done++;
  }

  return done / ((float) manager->itemCount);
}

bool assetManagerIsItemFinished(assetmanageritem_t *item) {
  // Sync done is always done
  if(item->state == ASSET_MANAGER_STATE_SYNC_DONE) return true;
  // Only check if ASYNC is done.
  if(item->state != ASSET_MANAGER_STATE_ASYNC_DONE) return false;
  // Does it need to still sync load?
  if(ASSET_MANAGER_LOADERS[item->type].loadSync == NULL) return true;
  return false;
}


// Thread Management
int32_t _assetManagerThread(thread_t *thread) {
  // TODO: Can I allow multiple threads to run?
  uint8_t i;
  assetmanager_t *manager;
  assetmanageritem_t *item;
  assetmanagerloaderdefinition_t *definition;
  bool result;
  manager = thread->user;

  for(i = 0; i < manager->itemCount; i++) {
    item = manager->items + i;
    definition = ASSET_MANAGER_LOADERS + item->type;

    // Only bother with ASYNC items
    if(definition->loadAsync == NULL) continue;

    // Are we already loading or already tried to load?
    if(item->state != ASSET_MANAGER_STATE_PENDING) continue;

    // Begin loading
    item->state = ASSET_MANAGER_STATE_ASYNC_LOADING;
    result = definition->loadAsync(item);

    // Finish Loading
    if(!result) {
      item->state = ASSET_MANAGER_STATE_ASYNC_ERROR;
    } else {
      item->state = ASSET_MANAGER_STATE_ASYNC_DONE;
    }
  }

  manager->finished = assetManagerProgressGet(manager) >= 1.0f;

  return 0;
}

void assetManagerUpdate(assetmanager_t *manager) {
  uint8_t i;
  assetmanageritem_t *item;
  assetmanagerloaderdefinition_t *definition;
  bool result;

  for(i = 0; i < manager->itemCount; i++) {
    item = manager->items + i;
    definition = ASSET_MANAGER_LOADERS + item->type;

    // If requires ASYNC loading, then confirm it has finished loading.
    if(definition->loadAsync != NULL) {
      if(item->state != ASSET_MANAGER_STATE_ASYNC_DONE) continue;
    } else if(item->state != ASSET_MANAGER_STATE_PENDING) {
      continue;
    } else if(definition->loadSync == NULL) {
      continue;
    }
    
    // Begin sync loading
    item->state = ASSET_MANAGER_STATE_SYNC_LOADING;
    result = definition->loadSync(item);

    // Finish loading
    if(!result) {
      item->state = ASSET_MANAGER_STATE_SYNC_ERROR;
    } else {
      item->state = ASSET_MANAGER_STATE_SYNC_DONE;
    }
  }

  manager->finished = assetManagerProgressGet(manager) >= 1.0f;
}

assetmanageritem_t * assetManagerItemAdd(assetmanager_t *manager) {
  assetmanageritem_t *item = manager->items + manager->itemCount++;
  item->state = ASSET_MANAGER_STATE_PENDING;
  return item;
}


// Texture
assetmanageritem_t * assetManagerLoadTexture(
  assetmanager_t *manager, texture_t *texture, char *fileName
) {
  assetmanageritem_t *item = assetManagerItemAdd(manager);
  
  item->type = ASSET_MANAGER_TYPE_TEXTURE;
  item->data.texture.fileName = fileName;
  item->data.texture.texture = texture;
  
  return item;
}

bool _assetManagerLoaderTextureAsync(assetmanageritem_t *item) {
  assetbuffer_t *buffer;
  int channels;
  stbi_io_callbacks OPENGL_STBI_CALLBACKS;

  buffer = assetBufferOpen(item->data.texture.fileName);
  if(buffer == NULL) return false;

  // Setup the interface for STBI
  OPENGL_STBI_CALLBACKS.read = &assetBufferRead;
  OPENGL_STBI_CALLBACKS.skip = &assetBufferSkip;
  OPENGL_STBI_CALLBACKS.eof = &assetBufferEnd; 

  // Buffer the image
  channels = 0;
  item->data.texture.data = (pixel_t *)stbi_load_from_callbacks(
    &OPENGL_STBI_CALLBACKS, buffer,
    &item->data.texture.width, &item->data.texture.height,
    &channels, STBI_rgb_alpha
  );

  // Close the buffer
  assetBufferClose(buffer);
  if(item->data.texture.data == NULL) return false;
  return true;
}

bool _assetManagerLoaderTextureSync(assetmanageritem_t *item) {
  // Turn into a texture.
  textureInit(
    item->data.texture.texture,
    item->data.texture.width,
    item->data.texture.height,
    item->data.texture.data
  );

  // Cleanup
  stbi_image_free(item->data.texture.data);
  return true;
}


// Font
assetmanageritem_t * assetManagerLoadFont(
  assetmanager_t *manager, font_t *font, char *fileName
) {
  assetmanageritem_t *item = assetManagerItemAdd(manager);
  
  item->type = ASSET_MANAGER_TYPE_FONT;
  item->data.font.fileName = fileName;
  item->data.font.font = font;

  return item;
}

bool _assetManagerLoaderFontAsync(assetmanageritem_t *item) {
  item->data.font.data = assetStringLoad(item->data.font.fileName);
  return item->data.font.data != NULL;
}

bool _assetManagerLoaderFontSync(assetmanageritem_t *item) {
  fontInit(item->data.font.font, item->data.font.data);
  free(item->data.font.data);
  return true;
}


// Shader
assetmanageritem_t * assetManagerShaderLoad(
  assetmanager_t *manager, shader_t *shader, char *fileVert, char *fileFrag
) {
  assetmanageritem_t *item = assetManagerItemAdd(manager);

  item->type = ASSET_MANAGER_TYPE_SHADER;
  item->data.shader.shader = shader;
  item->data.shader.fileVert = fileVert;
  item->data.shader.fileFrag = fileFrag;

  return item;
}

bool _assetManagerLoaderShaderAsync(assetmanageritem_t *item) {
  item->data.shader.dataVert = assetStringLoad(item->data.shader.fileVert);
  if(item->data.shader.dataVert == NULL) return false;

  item->data.shader.dataFrag = assetStringLoad(item->data.shader.fileFrag);
  if(item->data.shader.dataFrag == NULL) {
    free(item->data.shader.fileVert);
    return false;
  }
  
  return true;
}

bool _assetManagerLoaderShaderSync(assetmanageritem_t *item) {
  shaderInit(
    item->data.shader.shader,
    item->data.shader.dataVert,
    item->data.shader.dataFrag
  );
  free(item->data.shader.dataFrag);
  free(item->data.shader.dataVert);
  return true;
}